﻿-------------------------------------------------------------------------------
-- Are you local?
-------------------------------------------------------------------------------

local tablet = AceLibrary("Tablet-2.0")
local dewdrop = AceLibrary("Dewdrop-2.0")
local deformatter = AceLibrary("Deformat-2.0")
local crayon = AceLibrary("Crayon-2.0")
local BC = AceLibrary("Babble-Class-2.2")
local L = AceLibrary("AceLocale-2.2"):new("GroupFu")
local roster = nil

local _G = getfenv(0)

local masterLooter = nil
local rollers = nil
local lastAnnouncement = nil

local currentMin = 0
local currentMax = 0

-------------------------------------------------------------------------------
-- Addon Declaration
-------------------------------------------------------------------------------

GroupFu = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0", "FuBarPlugin-2.0")
local GroupFu = GroupFu

local options = {
	type = "group",
	args = {
		announce = {
			type = "toggle",
			name = L["Announce"],
			desc = L["Announce count down and display winner when roll clearing timer reached"],
			get = function() return GroupFu.db.profile.AnnounceRollCountdown end,
			set = function(v) GroupFu.db.profile.AnnounceRollCountdown = v end,
		},
		standard = {
			type = "toggle",
			name = L["Only accept 1-100"],
			desc = L["Accept standard (1-100) rolls only."],
			get = function() return GroupFu.db.profile.StandardRollsOnly end,
			set = function(v) GroupFu.db.profile.StandardRollsOnly = v end,
		},
		click = {
			type = "toggle",
			name = L["Perform roll when clicked"],
			desc = L["Perform a standard 1-100 roll when the FuBar plugin is clicked."],
			get = function() return GroupFu.db.profile.RollOnClick end,
			set = function(v) GroupFu.db.profile.RollOnClick = v end,
		},
		announcelocation = {
			type = "text",
			name = L["Announce location"],
			desc = L["Where to output roll results."],
			get = function() return GroupFu.db.profile.OutputChannel end,
			set = function(v) GroupFu.db.profile.OutputChannel = v end,
			validate = {
				["AUTO"] = L["Auto (based on group type)"],
				["LOCAL"] = L["Local"],
				["SAY"] = L["Say"],
				["PARTY"] = L["Party"],
				["RAID"] = L["Raid"],
				["GUILD"] = L["Guild"]
			}
		},
		rollclearing = {
			type = "text",
			name = L["Roll clearing"],
			desc = L["When to clear the rolls."],
			get = function() return GroupFu.db.profile.ClearTimer end,
			set = function(v) GroupFu.db.profile.ClearTimer = v end,
			validate = {
				["0"] = L["Never"],
				["15"] = L["15 seconds"],
				["30"] = L["30 seconds"],
				["45"] = L["45 seconds"],
				["60"] = L["60 seconds"]
			},
		},
		loottype = {
			type = "text",
			name = L["Loot type"],
			desc = L["Set the loot type."],
			get = function() return GetLootMethod() end,
			set = function(v)
				if v == "master" then
					SetLootMethod(v, UnitName("player"))
				else
					SetLootMethod(v)
				end
				dewdrop:Close()
				GroupFu:UpdateDisplay()
			end,
			validate = {
				["group"] = L["group"],
				["master"] = L["master"],
				["freeforall"] = L["freeforall"],
				["roundrobin"] = L["roundrobin"],
				["needbeforegreed"] = L["needbeforegreed"],
			},
			disabled = function()
				if UnitExists("party1") or UnitInRaid("player") and IsPartyLeader() or IsRaidLeader() then
					return false
				end
				return true
			end,
		},
		lootthreshold = {
			type = "text",
			name = L["Loot threshold"],
			desc = L["Set the loot threshold."],
			get = function() return tostring(GetLootThreshold()) end,
			set = function(v)
				SetLootThreshold(tonumber(v))
				dewdrop:Close()
				GroupFu:UpdateDisplay()
			end,
			validate = {
				["0"] = _G["ITEM_QUALITY0_DESC"],
				["1"] = _G["ITEM_QUALITY1_DESC"],
				["2"] = _G["ITEM_QUALITY2_DESC"],
				["3"] = _G["ITEM_QUALITY3_DESC"],
				["4"] = _G["ITEM_QUALITY4_DESC"],
				["5"] = _G["ITEM_QUALITY5_DESC"],
				["6"] = _G["ITEM_QUALITY6_DESC"],
			},
			disabled = function()
				if UnitExists("party1") or UnitInRaid("player") and IsPartyLeader() or IsRaidLeader() then
					return false
				end
				return true
			end,
		},
	}
}

-------------------------------------------------------------------------------
-- Initialization
-------------------------------------------------------------------------------

function GroupFu:OnInitialize()
	local revision = tonumber((string.match("$Revision: 27461 $", "^.-(%d+).-$"))) or 1
	if not self.version then self.version = "1" end
	self.version = self.version .. "." .. revision
	self.revision = revision

	_G.FUBAR_GROUPFU_VERSION = self.version
	_G.FUBAR_GROUPFU_REVISION = self.revision

	self.hasIcon = "Interface\\Buttons\\UI-GroupLoot-Dice-Up"
	self.clickableTooltip = false
	self.OnMenuRequest = options

	self:RegisterDB("GroupFuDB")
	self:RegisterDefaults("profile", {
		RollOnClick = true,
		OutputChannel = "PARTY",
		ClearTimer = "30",
		StandardRollsOnly = true,
		AnnounceRollCountdown = false,
		TextMode = "GROUPFU",
	})
end

function GroupFu:OnEnable()
	self:RegisterEvent("CHAT_MSG_SYSTEM")
	self:RegisterEvent("RAID_ROSTER_UPDATE", "UpdateDisplay")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED", "UpdateDisplay")
	self:RegisterEvent("PARTY_LOOT_METHOD_CHANGED", "UpdateDisplay")

	if AceLibrary:HasInstance("Roster-2.1") then roster = AceLibrary("Roster-2.1") end
end

function GroupFu:GetHighestRoller()
	local highestPlayer = nil
	for i, v in ipairs(rollers) do
		if not highestPlayer then
			highestPlayer = i
		else
			if v.roll > rollers[highestPlayer].roll then
				highestPlayer = i
			end
		end
	end
	return highestPlayer
end

function GroupFu:OnTextUpdate()
	if rollers then
		local num = UnitInRaid("player") and GetNumRaidMembers() or GetNumPartyMembers() + 1
		local highest = self:GetHighestRoller()
		local playerText = string.format(L["%s [%s]"], rollers[highest].player, "|cff"..crayon:GetThresholdHexColor(rollers[highest].roll, 0, 100)..rollers[highest].roll.."|r")
		self:SetText( string.format(L["%s (%d/%d)"], playerText, #rollers, num) )
	else
		self:SetText(self:GetLootTypeText())
	end
end

local function RollSorter(a, b)
	return a.roll > b.roll
end

function GroupFu:OnTooltipUpdate()
	local cat = tablet:AddCategory("columns", 2)
	local inRaid = UnitInRaid("player")
	cat:AddLine("text", L["Loot method"], "text2", self:GetLootTypeText())
	if inRaid and roster then
		if not masterLooter and GetLootMethod() == "master" then
			for i = 1, 40 do
				local ml = select(11, GetRaidRosterInfo(i))
				if ml then
					masterLooter = GetRaidRosterInfo(i)
					break
				end
			end
		end
		if masterLooter then
			local color = BC:GetHexColor(roster:GetUnitObjectFromName(masterLooter).class)
			cat:AddLine("text", L["Master looter"], "text2", "|cff"..color..masterLooter)
		end
		local officers = ""
		for n, u in pairs(roster.roster) do
			if u.rank == 1 then
				if officers ~= "" then officers = officers .. ", " end
				officers = officers.."|cff"..BC:GetHexColor(u.class)..u.name.."|r"
			elseif u.rank == 2 then
				cat:AddLine("text", L["Leader"], "text2", "|cff"..BC:GetHexColor(u.class)..u.name)
			end
		end
		if officers ~= "" then
			cat:AddLine("text", L["Officers"], "text2", officers)
		end
	end

	if rollers then
		cat = tablet:AddCategory(
			"text", L["Player"],
			"text2", L["Roll"],
			"columns", 2,
			"showWithoutChildren", false
		)
		table.sort(rollers, RollSorter)
		for i, v in ipairs(rollers) do
			local hexcolor = BC:GetHexColor(v.class)
			cat:AddLine(
				"text", string.format("|cff%s%s (%d %s)|r", hexcolor, v.player, v.level, v.class),
				"text2", "|cff"..crayon:GetThresholdHexColor(v.roll, 0, 100)..v.roll.."|r"
			)
		end
		local num = inRaid and GetNumRaidMembers() or GetNumPartyMembers() + 1
		cat:AddLine()
		cat:AddLine("text", string.format(L["%d of expected %d rolls recorded."], #rollers, num))
	end
	if self.db.profile.RollOnClick then
		tablet:SetHint(L["|cffeda55fClick|r to roll, |cffeda55fCtrl-Click|r to output winner, |cffeda55fShift-Click|r to clear the list."])
	else
		tablet:SetHint(L["|cffeda55fCtrl-Click|r to output winner, |cffeda55fShift-Click|r to clear the list."])
	end
end


function GroupFu:OnClick()
	if IsControlKeyDown() then
		self:AnnounceWinner()
	elseif IsShiftKeyDown() then
		self:ClearRolls()
	elseif self.db.profile.RollOnClick then
		RandomRoll("1", "100")
	end
end

function GroupFu:CHAT_MSG_SYSTEM(msg)
	-- Trap name of master looter if it has changed
	local mlname = deformatter(msg, ERR_NEW_LOOT_MASTER_S)
	if mlname then
		masterLooter = mlname
		self:UpdateDisplay()
		return
	end

	-- Trap rolls
	local player, roll, minRoll, maxRoll = deformatter(msg, RANDOM_ROLL_RESULT)
	if player then
		local db = self.db.profile

		roll = tonumber(roll)
		minRoll = tonumber(minRoll)
		maxRoll = tonumber(maxRoll)

		if not roll or not minRoll or not maxRoll
		or db.StandardRollsOnly and (minRoll ~= 1 or maxRoll ~= 100) then
			return
		elseif not db.StandardRollsOnly and rollers then
			-- If someone else has already rolled, and we accept rolls other
			-- than 1-100, assume that everyone should roll on the same premises
			-- and only accept rolls that have the same range as the first roll.
			if maxRoll ~= currentMax or minRoll ~= currentMin then
				return
			end
		end

		if not rollers then
			currentMin = minRoll
			currentMax = maxRoll
			rollers = {}
		end

		-- Ignore duplicate rolls.
		for i, v in ipairs(rollers) do
			if v.player == player then return end
		end

		local inRaid = UnitInRaid("player")
		local num = inRaid and GetNumRaidMembers() or GetNumPartyMembers() + 1

		if player == UnitName("player") then
			table.insert(rollers, {
				player = player,
				roll = roll,
				class = UnitClass("player"),
				level = UnitLevel("player"),
			})
		else
			for i = 1, num do
				local id = inRaid and "raid"..i or "party"..i
				if player == UnitName(id) then
					table.insert(rollers, {
						player = player,
						roll = roll,
						class = UnitClass(id),
						level = UnitLevel(id),
					})
				end
			end
		end

		if db.AnnounceRollCountdown or tonumber(db.ClearTimer) > 0 then
			-- If everyone has rolled, just output the winner.
			if num == #rollers and db.AnnounceRollCountdown then
				self:CancelScheduledEvent("GroupFuAnnounce")
				self:CancelScheduledEvent("GroupFuTimeout")
				self:AnnounceWinner()
			else
				self:CancelScheduledEvent("GroupFuAnnounce")
				self:ScheduleEvent("GroupFuTimeout", self.RollTimeout, tonumber(db.ClearTimer) - 5 or 5, self)
			end
		end

		self:UpdateDisplay()
	end
end

function GroupFu:ClearRolls()
	self:CancelScheduledEvent("GroupFuTimeout")
	self:CancelScheduledEvent("GroupFuAnnounce")
	if type(rollers) == "table" then
		for i, v in ipairs(rollers) do
			for k in pairs(v) do
				v[k] = nil
			end
			rollers[i] = nil
		end
	end
	rollers = nil
	self:Update()
end

function GroupFu:GetLootTypeText()
	if GetNumPartyMembers() > 0 or GetNumRaidMembers() > 0 then
		return ITEM_QUALITY_COLORS[GetLootThreshold()].hex .. L[GetLootMethod()]
	else
		return "|cff9d9d9d" .. L["Solo"]
	end
end

function GroupFu:RollTimeout()
	if not rollers then return end

	if self.db.profile.AnnounceRollCountdown then
		local num = UnitInRaid("player") and GetNumRaidMembers() or GetNumPartyMembers() + 1
		self:AnnounceOutput(string.format(L["Roll ending in 5 seconds, recorded %d of %d rolls."], #rollers, num))
	end
	self:ScheduleEvent("GroupFuAnnounce", self.AnnounceWinner, 5, self)
end

function GroupFu:AnnounceWinner()
	if self.db.profile.AnnounceRollCountdown then
		if rollers then
			local highest = self:GetHighestRoller()
			local tiedRollers = nil
			for i, v in ipairs(rollers) do
				if i ~= highest and v.roll == rollers[highest].roll then
					if not tiedRollers then tiedRollers = {} end
					table.insert(tiedRollers, i)
				end
			end
			if tiedRollers then
				table.insert(tiedRollers, highest)
				local playerNames = ""
				for i, v in ipairs(tiedRollers) do
					if playerNames == "" then
						playerNames = rollers[v].player
					else
						playerNames = playerNames .. ", " .. rollers[v].player
					end
				end
				lastAnnouncement = string.format(L["Tie: %s are tied at %d."], playerNames, rollers[highest].roll)
			else
				lastAnnouncement = string.format(L["Winner: %s."], string.format(L["%s [%s]"], rollers[highest].player, rollers[highest].roll))
			end
		end
		if not lastAnnouncement then return end
		self:AnnounceOutput(lastAnnouncement)
	end
	self:ClearRolls()
end

function GroupFu:AnnounceOutput( msg )
	if self.db.profile.OutputChannel == "LOCAL" then
		self:Print(msg)
	elseif self.db.profile.OutputChannel == "AUTO" then
		if GetNumRaidMembers() > 0 then
			SendChatMessage(msg, "RAID")
		elseif GetNumPartyMembers() > 0 then
			SendChatMessage(msg, "PARTY")
		else
			self:Print(msg)
		end
	else
		SendChatMessage(msg, self.db.profile.OutputChannel)
	end
end

